from dataclasses import dataclass
from injector import inject
from lmoe.api.lmoe_query import LmoeQuery
from lmoe.commands.print_classification import PrintClassification
from lmoe.commands.print_classifier_modelfile import PrintClassifierModelfile
from lmoe.commands.query import Query
from lmoe.commands.refresh import Refresh
from lmoe.commands.version import Version

import argparse


@inject
@dataclass
class CommandRunner:

    parsed_args: argparse.Namespace
    lmoe_query: LmoeQuery
    print_classification: PrintClassification
    print_classifier_modelfile: PrintClassifierModelfile
    query: Query
    refresh: Refresh
    version: Version

    def execute(self):
        command = None

        if self.parsed_args.classify:
            command = self.print_classification
        elif self.parsed_args.classifier_modelfile:
            command = self.print_classifier_modelfile
        elif self.parsed_args.refresh:
            command = self.refresh
        elif self.parsed_args.version:
            command = self.version
        else:
            command = self.query

        command.execute(self.parsed_args, self.lmoe_query)



from lmoe.api.lmoe_query import LmoeQuery
from abc import ABC, abstractmethod

import argparse


class Command(ABC):

    @abstractmethod
    def execute(self, parsed_args: argparse.Namespace, lmoe_query: LmoeQuery) -> None:
        pass



from lmoe.api.lmoe_query import LmoeQuery
from lmoe.commands.command import Command
from lmoe.definitions import get_project_version

import argparse


class Version(Command):

    def execute(self, parsed_args: argparse.Namespace, lmoe_query: LmoeQuery) -> None:
        print(get_project_version())



import importlib.resources
import os
import toml
import pathlib


def get_template_module():
    """The module string within which all of the template resource files reside."""
    return "lmoe.templates"


ABSOLUTE_PYPROJECT_FILENAME = pathlib.Path(__file__).parent.parent / "pyproject.toml"


def get_pyproject_filename():
    return ABSOLUTE_PYPROJECT_FILENAME


# Cache data for the root pyproject.toml file.
_PYPROJECT_DATA = None


def get_project_data():
    """Return the pyproject.toml file structured as a multi-layer dict.

    Caches loaded data so that this is only retrieved from the filesystem once.
    """
    global _PYPROJECT_DATA
    if _PYPROJECT_DATA is None:
        with open(ABSOLUTE_PYPROJECT_FILENAME, "r") as file:
            _PYPROJECT_DATA = toml.load(file)
    return _PYPROJECT_DATA


def get_project_version():
    """Returns the semantic versioning string associated with this project."""
    return get_project_data()["tool"]["poetry"]["version"]


# The absolute filenames of all resource template files
_TEMPLATE_FILE_NAMES = None


def get_template_file_names():
    global _TEMPLATE_FILE_NAMES
    if _TEMPLATE_FILE_NAMES is None:
        template_dir_object = importlib.resources.files(get_template_module())
        _TEMPLATE_FILE_NAMES = [
            os.path.join(template_dir_object.name, file_object.name)
            for file_object in template_dir_object.iterdir()
        ]
    return _TEMPLATE_FILE_NAMES



from lmoe.api.base_expert import BaseExpert
from lmoe.api.lmoe_query import LmoeQuery
from lmoe.framework.expert_registry import expert
from lmoe.utils.templates import (
    Example,
    read_template,
    read_yaml_examples,
    render_examples,
)
from string import Template

import ollama


# @expert
class Personality(BaseExpert):

    def __init__(self):
        self.examples = read_yaml_examples("personality.examples.yaml")

    @classmethod
    def name(cls):
        return "PERSONALITY"

    @classmethod
    def has_modelfile(cls):
        return True

    def description(self):
        return "A personality generator, who will use general knowledge and information provided by the user to answer their queries in a fun way, implementing a fun personality for the character Lmoe Armadillo."

    def example_queries(self):
        return [example.user_query for example in self.examples]

    def modelfile_contents(self):
        modelfile_template = Template(read_template(self.modelfile_name()))
        return modelfile_template.substitute(
            all_examples=render_examples(self.examples)
        )

    def generate(self, lmoe_query: LmoeQuery):
        stream = ollama.generate(
            model="lmoe_personality",
            prompt=lmoe_query.render(),
            stream=True,
        )
        for chunk in stream:
            if chunk["response"] or not chunk["done"]:
                print(chunk["response"], end="", flush=True)
        print("")



from dataclasses import dataclass
from injector import inject
from lmoe.api.lmoe_query import LmoeQuery
from lmoe.commands.command import Command
from lmoe.experts.refresh import Refresh

import argparse


@inject
@dataclass
class Refresh(Command):

    refresh: Refresh

    def execute(self, parsed_args: argparse.Namespace, lmoe_query: LmoeQuery) -> None:
        self.refresh.generate(lmoe_query)



import importlib.resources
import yaml

from dataclasses import dataclass
from lmoe.definitions import get_template_module
from string import Template
from typing import List, Optional

# Mapping from filename to Traversable object
_TEMPLATES = None


def load_templates():
    global _TEMPLATES
    if _TEMPLATES is None:
        template_dir_object = importlib.resources.files(get_template_module())
        _TEMPLATES = {
            file_object.name: file_object
            for file_object in template_dir_object.iterdir()
        }
    return _TEMPLATES


def read_template(filename):
    templates = load_templates()
    if not filename in templates:
        raise Exception("Could not find: " + filename)
    template = templates[filename]
    return template.read_text()


def read_yaml_file(filename):
    templates = load_templates()
    if not filename in templates:
        raise Exception("Could not find: " + filename)
    yaml_file = templates[filename]
    return yaml.safe_load(yaml_file.read_text())


@dataclass
class Example:
    """Example user context, query, and response."""

    stdin_context: Optional[str]
    paste_context: Optional[str]
    user_query: Optional[str]
    agent_response: Optional[str]


def read_yaml_examples(filename):
    """Reads a yaml file as a list of lmoe.utils.templates.Example instances."""
    return [Example(**example_dict) for example_dict in read_yaml_file(filename)]


_EXAMPLE_TEMPLATE = Template(
    """
Example $example_index)

===stdin-context===
$stdin_context
===stdin-context===
===paste-context===
$paste_context
===paste-context===
===user-query===
$user_query
===user-query===
$agent_response
"""
)


def render_examples(examples) -> str:
    rendered_examples = []
    example_index = 0
    for example in examples:
        example_index += 1
        rendered_examples.append(
            _EXAMPLE_TEMPLATE.substitute(
                example_index=example_index,
                stdin_context=example.stdin_context,
                paste_context=example.paste_context,
                user_query=example.user_query,
                agent_response=example.agent_response,
            )
        )
    return "\n\n".join(rendered_examples)



import os
import pathlib
import re

from lmoe.definitions import get_project_version


def update_readme_version():
    # Get project version
    project_version = get_project_version()

    # Update README.md in the project root
    readme_path = pathlib.Path(__file__).parent.parent / "README.md"
    update_version_in_file(readme_path, project_version)

    # Update README.md in the docs directory
    docs_readme_path = pathlib.Path(__file__).parent.parent / os.path.join(
        "docs", "README.md"
    )
    update_version_in_file(docs_readme_path, project_version)


def update_version_in_file(file_path, version):
    with open(file_path, "r") as file:
        content = file.read()

    # Replace version string
    updated_content = re.sub(
        r"Version[0-9]+\.[0-9]+\.[0-9]+", f"Version {version}", content
    )

    with open(file_path, "w") as file:
        file.write(updated_content)

    print(f"Updated {file_path} with version {version}.")



from lmoe.api.base_expert import BaseExpert
from lmoe.api.lmoe_query import LmoeQuery
from lmoe.framework.expert_registry import expert

import ollama


@expert
class ProjectInitialization(BaseExpert):

    @classmethod
    def name(cls):
        return "PROJECT_INITIALIZATION"

    @classmethod
    def has_modelfile(cls):
        return True

    def description(self):
        return "A model designed to initialize programming projects. Given verbal and ascii art descriptions and context, initializes programming projects in a number of different languages. Does not answer questions about how to initialize, only commands to perform initialization."

    def example_queries(self):
        return [
            "create a project for a java webserver using spring to serve the contents of the '/web/ directory",
            "initialize a project for a new npm module",
            "create a project like this",
            "make a project like this one",
        ]

    def generate(self, lmoe_query: LmoeQuery):
        stream = ollama.generate(
            model="lmoe_project_initialization",
            prompt=lmoe_query.render(),
            stream=True,
        )
        for chunk in stream:
            if chunk["response"] or not chunk["done"]:
                print(chunk["response"], end="", flush=True)
        print("")



from injector import Injector
from lmoe.commands.command_runner import CommandRunner
from lmoe.framework.native_module import NativeModule
from lmoe.framework.plugin_module_registry import PluginModuleRegistry

import importlib.util
import lmoe.experts
import os
import pkgutil
import sys


def import_experts(package_path):
    for file_name in os.listdir(package_path):
        if file_name.endswith(".py") and file_name != "__init__.py":
            module_name = file_name[:-3]  # Remove '.py' extension
            module_path = os.path.join(package_path, file_name)

            spec = importlib.util.spec_from_file_location(module_name, module_path)
            if spec is not None:
                module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(module)


def run():
    # Dynamically import native experts
    native_path = os.path.dirname(lmoe.experts.__file__)
    import_experts(native_path)

    # Dynamically import plugin experts
    # TODO: support for configuring the plugin directory, naming, etc.
    home_directory = os.path.expanduser("~")

    # Add the parent plugin directory to the Python path
    plugin_parent_path = f"{home_directory}/lmoe_plugins"
    sys.path.append(plugin_parent_path)

    # Import the plugin Python module
    plugin_path = f"{plugin_parent_path}/lmoe_plugins"
    import_experts(plugin_path)

    # Carry on with app instantiation, install native and plugin injection modules
    modules = [NativeModule]
    modules.extend(PluginModuleRegistry.modules())
    injector = Injector(modules)

    # Defer execution to the command runner
    command_runner = injector.get(CommandRunner)
    command_runner.execute()



from lmoe.api.base_expert import BaseExpert
from lmoe.api.lmoe_query import LmoeQuery
from lmoe.framework.expert_registry import expert

import base64
import binascii
import ollama
import re
import requests


def extract_base64(input_string):
    # Use regular expression to extract Base64 data
    match = re.search(r"\s*(user_context:)?\s*([A-Za-z0-9+/]+={0,2})", input_string)
    if match:
        base64_data = match.group(2)
        # Remove whitespace and padding characters
        base64_data = base64_data.replace(r"[A-Za-z0-9+/=]", "")
        # Add padding if necessary
        while len(base64_data) % 4 != 0:
            base64_data += "="
        return base64_data
    else:
        return None


def check_base64(s):
    try:
        if not s.isascii():
            print("s is not ascii")
            return False
        decoded_bytes = base64.standard_b64decode(s)
        decoded_string = base64.standard_b64encode(decoded_bytes)
        return True
    except Exception as e:
        print(e)
        return False


@expert
class Image(BaseExpert):

    @classmethod
    def name(cls):
        return "IMAGE"

    @classmethod
    def has_modelfile(cls):
        return True

    def description(self):
        return "Analyzes the contents of images and answers questions about them."

    def example_queries(self):
        return [
            "what's in this picture",
            "what's in this iamge",
            "how many people are in this picture",
            "where was this picture taken",
        ]

    def generate(self, lmoe_query: LmoeQuery):
        # TODO: Have a good way to get image context (interpret base64 context, then just from filename)
        extracted_base64 = extract_base64(lmoe_query.stdin_context)
        if not check_base64(extracted_base64):
            print("bad base64")
            exit(1)

        # TODO: Ollama client for this is not working but raw HTTP calls are currently. Open a bug.
        # TODO: replace with a more general ollama client allowing configuration?
        # TODO: extract JSON, support streaming
        try:
            response = requests.post(
                "http://127.0.0.1:11434/api/generate",
                json={
                    "model": "llava",
                    "prompt": lmoe_query.user_query,
                    "stream": False,
                    "images": [extracted_base64],
                },
            )
            print(response.text)
        except requests.RequestException as e:
            print("Error:", e)
